import pickle
from itertools import combinations_with_replacement as comb_w_r

import numpy as np
import tensorflow as tf


class TensorMinMax:
    """Copy of sklearn's MinMaxScaler implemented to work with tensorflow.

    When used, tensorflow is able to take gradients on the transformation as
    well as on the network itself, allowing for gradient-based optimization in
    inverse design problems.

    Parameters
    ----------
        feature_range : 2-tuple, optional
            Desired range of transformed data. Defaults to (0, 1)
        copy : bool, optional
            Set to false to perform inplace operations. Defaults to True.
    """

    def __init__(self, feature_range=(0, 1), copy=True):
        self.feature_range = feature_range
        self.copy = copy
        self.min_ = None
        self.scale_ = None
        self.data_min = None
        self.data_max = None

    def fit(self, X):
        """Fits the transfomer to the data.

        Essentially finds original min and max of data to be able to shift the data.

        Parameters
        ----------
        X : tensor or ndarray
            Data to fit
        """
        self.data_min = np.amin(X, axis=0)
        self.data_max = np.amax(X, axis=0)
        self.scale_ = (self.feature_range[1] - self.feature_range[0]) / (
            self.data_max - self.data_min
        )
        self.min_ = self.feature_range[0] - self.data_min * self.scale_

    def transform(self, X, mode="numpy"):
        """Actually does the transorm.

        Parameters
        ----------
        X : tensor or ndarray
            Data to transform
        mode : {'numpy' or 'tensor'}, optional
            Whether to use numpy or tensorflow operations.

        Returns
        -------
        X : tensor or ndarray
            Transformed data
        """
        if mode == "numpy":
            X *= self.scale_
            X += self.min_
        elif mode == "tensor":
            X = X * tf.constant(self.scale_, tf.float32) + tf.constant(
                self.min_, tf.float32
            )

        return X

    def inverse_transform(self, X, mode="numpy"):
        """Undo the transorm.

        Parameters
        ----------
        X : tensor or ndarray
            Data to inverse transform
        mode : {'numpy' or 'tensor'}, optional
            Whether to use numpy or tensorflow operations.

        Returns
        -------
        X : tensor or ndarray
            Inverse transformed data
        """
        if mode == "numpy":
            X -= self.min_
            X /= self.scale_
        elif mode == "tensor":
            X = (X - tf.constant(self.min_, tf.float32)) / tf.constant(
                self.scale_, tf.float32
            )

        return X


class ImportNN:
    """Class to import trained NN.

    This the way we've been saving and using our neural networks. After saving them
    we can simply import them using this class and it keeps them open for as many
    operations as we desire.

    Attributes
    ----------
        normX : TensorMinMax
            Norm of the inputs
        normY: TensorMinMax)
            Norm of the outputs
        s_data : 2-tuple
            Dimensions (size) of input and outputs

    Parameters
    ----------
        directory : str
            The directory where the model has been stored
    """

    def __init__(self, directory):
        # import all graph info

        with open(f"{directory}/Import.pkl", "rb") as file:
            dict_ = pickle.load(file)
            self.normX = dict_["normX"]
            self.normY = dict_["normY"]
            self.s_data = dict_["s_data"]

        self.graph = tf.Graph()
        self.sess = tf.compat.v1.Session(graph=self.graph)
        with self.graph.as_default():
            # Import graph
            imported_meta = tf.compat.v1.train.import_meta_graph(f"{directory}/model.meta")
            imported_meta.restore(self.sess, f"{directory}/model")

            # get all tensor names
            self.output_tf = self.graph.get_tensor_by_name("OUTPUT:0")
            self.input_tf = self.graph.get_tensor_by_name("INPUT:0")
            self.input_tf_parts = [self.graph.get_tensor_by_name(f"INPUT_{i}:0") for i in range(self.s_data[0])]

            self.keep_prob = self.graph.get_tensor_by_name("KEEP_PROB:0")
        tf.compat.v1.disable_eager_execution()

    def validate_input(self, input):
        """Used to check for valid input.

        If it is only a single data point, expands the dimensions so it fits properly

        Parameters
        -----------
        input : ndarray
            Numpy array with width s_data[0] (hopefully)

        Returns
        --------
        input : ndarray
            Numpy array with width s_data[0] (hopefully) and height 1
        """
        # validate and prepare data
        input = np.array(input)
        # make sure it's 2-dimensional
        if len(input.shape) == 1:
            input = np.expand_dims(input, axis=1).T
        # make sure it's the right size
        if input.shape[1] != self.s_data[0]:
            raise ValueError("Data is the wrong size")

        return input

    def output(self, input, kp=1):
        """Runs input through neural network.

        Parameters
        ----------
        input : ndarray
            Numpy array with width s_data[0]
        kp : int, optional
                Value from 0 to 1, 1 refers to not performing any dropout on nodes, 0 drops all of them. Defaults to 1.

        Returns
        ----------
        output: ndarray
            numpy array with width s_data[1]
        """
        # validate data
        input = self.validate_input(input)
        # return the outputs
        return self.sess.run(
            self.normY.inverse_transform(self.output_tf),
            feed_dict={self.input_tf: input, self.keep_prob: kp},
        )

    def differentiate(self, input, d, kp=1):
        """Returns partial derivative of neural network.

        Parameters
        ----------
        input : ndarray
            numpy array with width s_data[0]
        d : 3-tuple of ints
            Refers to partial of first element wrt second element to the order of third element
        kp : int, optional
            Value from 0 to 1, 1 refers to not performing any dropout on nodes, 0 drops all of them. Defaults to 1.

        Returns
        ----------
            output : ndarray
                numpy array with width s_data[1]
        """
        # validate data
        input = self.validate_input(input)
        # make feed dict
        fd = {self.keep_prob: kp}
        for i in range(self.s_data[0]):
            fd[self.input_tf_parts[i]] = input[:, i : i + 1]
        # take first derivatives, then the rest
        deriv = tf.gradients(
            self.normY.inverse_transform(self.output_tf)[:, d[0] : d[0] + 1],
            self.input_tf_parts[d[1]],
        )[0]
        for _ in range(1, d[2]):
            deriv = tf.gradients(deriv, self.input_tf_parts[d[1]])[0]

        return self.sess.run(deriv, feed_dict=fd)

    def rel_error(self, input, output, kp=1):
        """Returns relative error of network.

        Parameters
        ----------
        input : ndarray
            Numpy array with width s_data[0]
        output : ndarray
            Numpy array with width s_data[1]
        kp : int, optional
            Value from 0 to 1, 1 refers to not performing any dropout on nodes, 0 drops all of them. Defaults to 1.

        Returns
        ----------
        relative error : scalar
            The relative error of inputs/outputs
        """
        # validate data
        input = self.validate_input(input)
        # get output
        output_nn = self.output(input, kp)
        # get rid of any possible divide by 0's)
        mask = ~np.isin(output, 0)
        # make relative error
        re = np.abs((output[mask] - output_nn[mask]) / output[mask])
        return re.mean()


class ImportLR:
    """Class to import trained Linear Regression.

    To remove independence on sklearn and it's updates, we manually implement an sklearn
    Pipeline that includes (PolynomialFeatures, LinearRegression). We use the actual sklearn
    implementation to train, save the coefficients, and then proceed to implement it here.
    To see how to save a pipeline like above to be used here see SiPANN/LR/regress.py

    Attributes
    -----------
        coef_ : ndarray
            Linear Regression Coefficients
        degree_ : float
            Degree to be used in PolynomialFeatures.
        s_data : 2-tuple
            Dimensions of inputs and outputs

    Parameters
    ----------
        directory : str
            The directory where the model has been stored
    """

    def __init__(self, directory):
        # import all graph info
        with open(directory, "rb") as file:
            dict_ = pickle.load(file)
            self.coef_ = dict_["coef_"]
            self.degree_ = dict_["degree_"]
            self.s_data = dict_["s_data"]

    def make_combos(self, X):
        """Duplicates Polynomial Features.

        Takes in an input X, and makes all possibly combinations of it using
        polynomials of specified degree.

        Parameters
        -----------
        X : ndarray
            Numpy array of size (N, s_data[0])

        Returns
        --------
        polyCombos : ndarray
            Numpy array of size (N, )
        """
        combos = []
        for i in range(self.degree_ + 1):
            combos += list(comb_w_r(range(self.s_data[0]), i))

        # make matrix of all combinations
        n = len(X)
        polyCombos = np.ones((n, len(combos)))
        for j, c in enumerate(combos):
            if c == ():
                polyCombos[:, j] = 1
            else:
                for k in c:
                    polyCombos[:, j] *= X[:, k]

        return polyCombos

    def validate_input(self, input):
        """Used to check for valid input.

        If it is only a single data point, expands the dimensions so it fits properly

        Parameters
        -----------
        input : ndarray
            Numpy array with width s_data[0] (hopefully)

        Returns
        --------
        input : ndarray
            Numpy array with width s_data[0] (hopefully) and height 1
        """
        # validate and prepare data
        input = np.array(input)
        # make sure it's 2-dimensional
        if len(input.shape) == 1:
            input = np.expand_dims(input, axis=1).T
        # make sure it's the right size
        if input.shape[1] != self.s_data[0]:
            raise ValueError("Data is the wrong size")

        return input

    def predict(self, X):
        """Predict values.

        Runs X through Pipeline to make prediction

        Parameters
        -----------
        X : ndarray
            Numpy array of size (N, s_data[0])

        Returns
        --------
        polyCombos : ndarray
            Numpy array of size (N, )
        """
        X = self.validate_input(X)
        Xcombo = self.make_combos(X)
        return Xcombo @ (self.coef_.T)



import ctypes
import os

import matplotlib.pyplot as plt
import nlopt
import numpy as np
from numba import njit
from numba import vectorize
from numba.extending import get_cython_function_address
from scipy import special
from tables import ComplexCol
from tables import Float64Col
from tables import IsDescription
from tables import open_file
from tqdm import tqdm

from SiPANN import scee

addr = get_cython_function_address("scipy.special.cython_special", "binom")
functype = ctypes.CFUNCTYPE(ctypes.c_double, ctypes.c_double, ctypes.c_double)
binom_fn = functype(addr)

##################################################
###              HELPER FUNCTIONS              ###
###  used to help quickly define gap functions ###
##################################################


@njit
def bernstein_quick(n, j, t):
    """Quickly computes the jth bernstein polynomial for the basis of n+1
    polynomials.

    Parameters
    -----------
    n : int
        The number of elements minus one in the basis of berstein polynomials
    j : int
        The index of bernstein polynomial that needs to be computed
    t : float
        [0-1] the value at which to compute the polynomial

    Returns
    ----------
    test : float
        Result of computing the jth berstein polynomial at t
    """
    return binom_fn(n, j) * t ** j * (1 - t) ** (n - j)


def bezier_quick(g, length):
    """Computes the bezier curve for the evenly spaced control points with gaps
    g.

    Parameters
    ----------
    g :  ndarray
        Numpy array of size (n,) of gap values at each of the control points
    length :  float
        length of the coupler

    Returns
    ----------
    result : dict
        {'g': original control points, 'w': length of coupler, 'f': bezier curve function defining gap function, 'df': derivative of gap function, 'd2f': 2nd derivative of gap functions}
    """
    n = len(g) - 1
    return {
        "g": g,
        "w": length,
        "f": lambda t: np.sum(
            np.array(
                [(g[j]) * bernstein_quick(n, j, t / length) for j in range(len(g))]
            ),
            axis=0,
        ),
        "df": lambda t: np.sum(
            np.array(
                [
                    n
                    * (g[j])
                    * (
                        bernstein_quick(n - 1, j - 1, t / length)
                        - bernstein_quick(n - 1, j, t / length)
                    )
                    for j in range(len(g))
                ]
            ),
            axis=0,
        )
        / length,
        "d2f": lambda t: np.sum(
            np.array(
                [
                    n
                    * (n - 1)
                    * (g[j] / 2)
                    * (
                        bernstein_quick(n - 2, j - 2, t / length)
                        - 2 * bernstein_quick(n - 2, j - 1, t / length)
                        + bernstein_quick(n - 2, j, t / length)
                    )
                    for j in range(len(g))
                ]
            ),
            axis=0,
        )
        / length ** 2,
    }


class Optimization(IsDescription):
    """Class to save the data in h5 files if dataCollect=True."""

    width = Float64Col()
    thickness = Float64Col()
    wavelength = Float64Col()
    k = ComplexCol(16)
    t = ComplexCol(16)
    length = Float64Col()
    g1 = Float64Col()
    g2 = Float64Col()
    g3 = Float64Col()
    g4 = Float64Col()
    g5 = Float64Col()
    g6 = Float64Col()
    g7 = Float64Col()
    g8 = Float64Col()


#########################################################
###                 ACTUAL OPTIMIZING                 ###
###  this function does all of the heavy lifting here ###
#########################################################
def make_coupler(
    goalK=0.4,
    arrayK=None,
    waveSweep=np.linspace(1500, 1600, 4),
    gapN=16,
    algo=35,
    edgeN=8,
    plot=False,
    collectData=False,
    width=500,
    thickness=220,
    radius=5000,
    maxiter=None,
    verbose=0,
):
    """Optimizes output from a directional coupler defined by a bezier curve to
    a specified output magnitude.

    Parameters
    ----------
    goalK :  float
        [0-1] mandatory, unless using arrayK. Desired \|kappa\|^2 magnitude
    arrayK :  ndarray, optional
        Has to have size (2,). [0-1] can specify a \|kappa\|^2 magnitude at start and end of wavelength sweep. Defaults to None
    waveSweep :  ndarray, optional
        Sweep of wavelengths to evaluate objective function at. Defaults to ``np.linspace(1500,1600,4)``
    gapN :  int, optional
        Number of control points that can vary. Defaults to 16.
    algo :  int, optional
        Optimization algorithm that nlopt uses. Defaults to 35
    edgeN :  int, optional
        Number of control points on each edge that are fixed at gap of 1500 nm. Defaults to 8.
    plot :  bool, optional
        If True then optimization will plot the current coupler at each iteration with the control points. Defaults to False.
    collectData :  bool, optional
        Whether to collect data for couplers of each iteration (could be useful for machine learning and even faster design). Defaults to False.
    width :  float, optional
        Width of waveguides in nm. Defaults to 500.
    thickness :  float, optional
        Thickness of waveguides in nm. Defaults to 220.
    radius :  float, optional
        Radius of allowable curve in directional coupler in nm. Defaults to 5000.
    maxiter : int, optional
        The number of max iterations to run each of the gloabl and local optimization for. If None, doesn't apply. Defaults to None.
    verbose :  int, optional
        Amount of logging to output. If 0, none. If 1, tqdm bar. If 2, prints all information (can be cumbersome). Defaults to 0.

    Returns
    ----------
    coupler : GapFuncSymmetric
        The final coupler object from SCEE
    control_pts : ndarray
        The control points defining bezier curve for gap function (nm)
    length : ndarray
        The length of the coupler (nm)
    """

    # initial values for the optimizer to use and bounds for values
    couplingWidth = 20000
    couplingMin = 5000
    couplingMax = 100000
    gapMin = 0
    gapMax = 1500
    iter = [0]
    mseVals = []
    coupler = scee.GapFuncSymmetric(
        width, thickness, lambda x: gapMax, lambda x: 0, 0, couplingWidth
    )
    curvatureFunc = lambda x: 0
    localOpt = False
    waveN = len(waveSweep)
    waveStart = waveSweep[0]
    waveStop = waveSweep[-1]

    # set up progress bars
    if verbose == 1:
        if maxiter is not None:
            loop = tqdm(total=maxiter * 2, position=0, leave=True)
        else:
            loop = tqdm(total=float("inf"), position=0, leave=True)

    # varying values of goal coupling if desired
    if arrayK is not None:
        goalK = np.array(
            [arrayK[0] if k < waveN / 2 else arrayK[1] for k in range(waveN)]
        )

    # sweep of wavelength to optimize over
    # waveSweep = np.linspace(waveStart, waveStop, waveN)
    # dataPoints = {str(wave):{'g': [], 'k': [], 't': []} for wave in range(waveStart,waveStop+1)}

    # define plot for debugging and final result
    if plot:
        plt.ion()
    fig, axes = plt.subplots(2, 1)

    axes[0].set_ylim(-gapMax, gapMax + 100)
    scatter1 = axes[0].scatter(
        np.linspace(0, couplingWidth, gapN), np.zeros(gapN), label="control points"
    )
    (line1,) = axes[0].plot(
        np.linspace(0, couplingWidth, 500), np.zeros(500), label="gap function"
    )
    (line2,) = axes[1].plot(
        waveSweep,
        np.zeros_like(waveSweep),
        "o-",
        label="cross-port goal = " + str(goalK),
    )
    axes[1].set_ylim(-5, 0)
    axes[0].set_xlabel("coupling length (nm)")
    axes[0].set_ylabel("coupling gap (nm)")
    axes[1].set_xlabel("wavelength (nm)")
    axes[1].set_ylabel("error (dB)")
    plt.tight_layout()
    axes[0].legend(loc=8)
    axes[1].legend(loc=8)

    # file to collect data in
    if collectData:
        h5file = open_file(
            "./data/" + ("%.2f" % goalK).split(".")[1] + "/" + "data.h5",
            mode="w",
            title="Simulation Data",
        )
        group = h5file.create_group("/", "coupler", "Optimization")
        table = h5file.create_table(group, "optdata", Optimization, "Data")
        point = table.row

    def f(g, grad):
        """Optimization function.

        Parameters
        -----------
        x :  ndarray
            Numpy array of size ((gapN + 1,) of control points, first element is length
        grad : ndarray
            gradient of the optimization (not used)

        Returns
        ----------
        result: float
            MSE of power equation (-10 log (|kappa|^2/goalK))
        """

        iter[0] += 1

        # define optimization points and distance between them
        x = np.linspace(0, g[0], 2 * gapN + 2 * edgeN)
        # mirror gap control points for symmetry
        g_total = np.append(
            np.append(np.append(np.full(edgeN, gapMax), g[1:]), g[-1:0:-1]),
            np.full(edgeN, gapMax),
        )

        # get interpolation of waveguide
        gap = bezier_quick(g_total, g[0])
        coupler.update(gap=gap["f"], dgap=gap["df"], zmax=gap["w"])

        # get the cross coupling
        currK = coupler.predict((1, 4), waveSweep)

        # draw current waveguide and power if debug mode on
        if plot and iter[0] % 10000 != 0:
            scatter1.set_offsets(np.array([x, g_total]).T)
            dom = np.linspace(0, g[0], 500)
            axes[0].set_xlim(0, g[0])

            line1.set_xdata(dom)
            line1.set_ydata(gap["f"](dom))
            line2.set_ydata(-np.abs(-10 * np.log10((np.abs(currK) ** 2) / goalK)))
            fig.canvas.draw()
            fig.canvas.flush_events()
        elif plot:
            scatter1.set_visible(False)
            line1.set_visible(False)
            x_plot = np.linspace(0, g[0], 500)
            axes[0].set_xlim(0, g[0])
            axes[0].get_legend().remove()

            y_plot = gap["f"](x_plot) / 2
            fill1 = axes[0].fill_between(x_plot, -y_plot - width, -y_plot, color="k")
            fill2 = axes[0].fill_between(x_plot, y_plot, y_plot + width, color="k")
            line2.set_ydata(-np.abs(-10 * np.log10((np.abs(currK) ** 2) / goalK)))
            fig.canvas.draw()
            fig.canvas.flush_events()
            plt.waitforbuttonpress()

            line1.set_visible(True)
            fill1.set_visible(False)
            fill2.set_visible(False)
            scatter1.set_visible(True)
            axes[0].legend(loc=8)

        # get the mean squared error between current coupling and goal
        mse = np.sum((np.log10((np.abs(currK) ** 2) / goalK)) ** 2) / waveN
        mseVals.append(mse)

        # add penalty if gap function has value < 100 nm (SCEE doesn't perform well for coupling closer than this)
        mse += np.sum(gap["f"](np.linspace(0, g[0], 100)) < 100)

        # # print iteration progress
        if verbose == 2:
            print(f"MSE: {mse}")
            print(f"currK: {np.abs(currK)**2}")
            print(f"g: {g}")
            # if the optimization changed from global constrained to local optimization
            print("local optimization:", localOpt, "\n")
        elif verbose == 1:
            o = "LOCAL" if localOpt else "GLOBAL"
            loop.update(1)
            loop.set_description(
                f"{o}, MSE: {np.round(mse,4)}, Mean currK: {np.round((np.abs(currK)**2).mean(),4)}"
            )

        return mse

    def constraint(x, grad, radius=5000):
        """Constraint function on waveguide curvature to make sure coupler
        reduces bending loss.

        Parameters
        ----------
        x :   ndarray
            Numpy array of size ((gapN + 1,) of control points, first element is length
        grad : ndarray
            gradient of optimization (not used)
        radius : float
            radius of allowable curve

        Returns
        ----------
        result : (bool)
            whether iteration satisfies curvature
        """

        # The derivative of the waveguide is half the derivative of the gap function
        deriv = lambda x: coupler.dgap(x) / 2

        # domain doesn't include 0 or full length because of division by zero
        domain = np.linspace(0.002, x[0] - 0.002, 500)
        # curvature at each point in the domain
        curve = np.abs(curvatureFunc(domain) / ((1 + (deriv(domain) ** 2)) ** 1.5))

        # save data if constraints are satisfied and collectData=True
        # collects gap points, length, width, thickness, integer wavelengths in range, and outputs from coupler
        if collectData and not np.any(curve > 1 / radius):
            waves = range(waveStart, waveStop + 1)
            currK = coupler.predict((1, 4), waves)
            currT = coupler.predict((1, 3), waves)
            for i in range(101):
                point["width"] = width
                point["thickness"] = thickness
                point["wavelength"] = waves[i]
                point["length"] = x[0]
                point["k"] = currK[i]
                point["t"] = currT[i]
                point["g1"] = x[1]
                point["g2"] = x[2]
                point["g3"] = x[3]
                point["g4"] = x[4]
                point["g5"] = x[5]
                point["g6"] = x[6]
                point["g7"] = x[7]
                point["g8"] = x[8]
                point.append()
                table.flush()

        return float(np.any(curve > 1 / radius))

    ##### initialize and perform global optimization  ########
    x0 = np.append(couplingWidth, np.linspace(gapMin, gapMax, gapN))
    opt = nlopt.opt(int(algo), gapN + 1)
    opt.set_lower_bounds(np.append(couplingMin, np.full((gapN), gapMin)))
    opt.set_upper_bounds(np.append(couplingMax, np.full((gapN), gapMax)))

    opt.set_min_objective(f)

    opt.set_xtol_rel(1e-5)
    opt.set_ftol_abs(1e-7)
    if maxiter is not None:
        opt.set_maxeval(maxiter)

    opt.add_equality_constraint(lambda x, grad: constraint(x, grad, radius), 1e-8)

    gap_final = opt.optimize(x0)

    ###### local optimize after global ########
    # set color to black so in debug mode we can see when the local optimization starts
    line1.set_color("k")
    scatter1.set_color("k")
    line2.set_color("k")
    localOpt = True

    opt = nlopt.opt(40, gapN + 1)

    # add constraint only on local optimization
    opt.add_equality_constraint(lambda x, grad: constraint(x, grad), 1e-8)

    # initialize and perform local optimization
    opt.set_lower_bounds(np.append(couplingMin, np.full((gapN), gapMin)))
    opt.set_upper_bounds(np.append(couplingMax, np.full((gapN), gapMax)))

    opt.set_min_objective(f)

    opt.set_xtol_rel(1e-5)
    opt.set_ftol_abs(1e-7)
    opt.set_ftol_rel(1e-6)
    if maxiter is not None:
        opt.set_maxeval(maxiter)

    gap_final = opt.optimize(gap_final)

    # close data collection file (h5 used so data is still collected if interrupted)
    if collectData:
        h5file.close()

    # make sure final coupler is defined properly
    gap_total = np.append(
        np.append(np.append(np.full(edgeN, gapMax), gap_final[1:]), gap_final[-1:0:-1]),
        np.full(edgeN, gapMax),
    )
    gap = bezier_quick(gap_total, gap_final[0])
    coupler.update(gap=gap["f"], dgap=gap["df"], zmax=gap["w"])

    # setup for final coupler plot
    if plot:
        dom = np.linspace(0, gap_final[0], 500)
        axes[0].set_xlim(0, gap_final[0])
        scatter1.set_visible(False)
    K_final = coupler.predict((1, 4), waveSweep)

    # plot optimized waveguide geometry and power
    if plot:
        line1.set_visible(False)
        line2.set_ydata(-np.abs(np.log10(np.abs(K_final) ** 2 / goalK)))
        axes[0].fill_between(
            dom, gap["f"](dom) / 2, gap["f"](dom) / 2 + width, color="k"
        )
        axes[0].fill_between(
            dom, -gap["f"](dom) / 2 - width, -gap["f"](dom) / 2, color="k"
        )
        axes[0].get_legend().remove()

        minf = opt.last_optimum_value()
        print(gap_final)
        print(minf, f(gap_final, None))
        print(np.abs(K_final) ** 2)
        fig.canvas.draw()
        fig.canvas.flush_events()
        # don't press a keyboard key once plot shows. If save or resize use mouse
        plt.waitforbuttonpress()

    # plot MSE over iterations
    if plot:
        fig, axes = plt.subplots(1, 1)
        axes.semilogy(mseVals)
        axes.set_xlabel("optimizer iterations")
        axes.set_ylabel("MSE")
        fig.canvas.draw()
        fig.canvas.flush_events()
        plt.waitforbuttonpress()

    # if collectData:
    #     for wave in range(waveStart,waveStop+1):
    #         np.savez('./data/' + ('%.2f'%goalK).split('.')[1] + '/' + str(wave),GAP=dataPoints[str(wave)]['g'],K=dataPoints[str(wave)]['k'],T=dataPoints[str(wave)]['t'])

    return coupler, gap_total, gap_final[0]


#########################################################
###              SAVING & LOADING COUPLERS            ###
###        useful for caching and reusing later       ###
#########################################################
def save_coupler(width, thickness, control_pts, length, filename):
    """Used to save optimized couplers efficiently.

    When used only saves gap points and coupling length into a .npz file. This allows for easy reloading
    via the functions below.

    Parameters
    ----------
    width :  float
        Width of the waveguide in nm
    thickness :  float
        Thickness of the waveguide in nm
    control_pts :  ndarray
        Gap points defining gap function via bernstein polynomials. Second parameter returned by ``make_coupler``.
    length :  float
        Length of corresponding coupler. Third parameter returned by ``make_coupler``
    filename :  string
        Name of file to write to.
    """
    np.savez(
        filename,
        WIDTH=np.array([width]),
        THICKNESS=np.array([thickness]),
        LENGTH=np.array([length]),
        GAP=control_pts,
    )


def load_coupler(filename):
    """Used to load optimized couplers efficiently.

    Any coupler saved using the ``save_coupler`` function can be reloaded using this one. It will return
    an instance of ``SiPANN.scee.GapFuncSymmetric``.

    Parameters
    ----------
    filename :  string
        Location where file is stored.

    Returns
    ----------
    coupler : GapFuncSymmetric
        Saved coupler
    length : float
        Length of coupler
    """
    # load all the data
    data = np.load(filename)
    g = data["GAP"]
    length = data["LENGTH"][0]
    width = data["WIDTH"][0]
    thickness = data["THICKNESS"][0]

    # make curve and return object
    bez = bezier_quick(g, length)
    return (
        scee.GapFuncSymmetric(width, thickness, bez["f"], bez["df"], 0, bez["w"]),
        length,
    )


def premade_coupler(split):
    """Loads premade couplers.

    Various splitting ratio couplers have been made and saved. This function reloads them. Note that each of their
    lengths are different and are also returned for the users info. These have all been designed with waveguide
    geometry 500nm x 220nm.

    Parameters
    ----------
    split :  int
        Percent of light coming out cross port. Valid numbers are 10, 20, 30, 40, 50, 100. 100 is a full crossover.

    Returns
    ----------
    coupler : GapFuncSymmetric
        Designed Coupler
    length : float
        Length of coupler
    """
    if split not in [10, 20, 30, 40, 50, 100]:
        raise ValueError("That splitting ratio hasn't been made")

    # load all the data
    filename = f"split_{split}_{100-split}.npz"
    data = np.load(
        os.path.join(os.path.dirname(os.path.realpath(__file__)), "COUPLER", filename)
    )
    g = data["GAP"]
    length = data["LENGTH"][0]
    width = data["WIDTH"][0]
    thickness = data["THICKNESS"][0]

    # make curve and return object
    bez = bezier_quick(g, length)
    return (
        scee.GapFuncSymmetric(width, thickness, bez["f"], bez["df"], 0, bez["w"]),
        length,
    )



import warnings
from abc import ABC
from abc import abstractmethod

import gdspy
import numpy as np
import pkg_resources
from scipy import special
from scipy.integrate import quad

from SiPANN.import_nn import ImportLR

##########################################################################################
####  We initialize all ANN's and regressions as global objects to speed things up.  #####
##########################################################################################
cross_file = pkg_resources.resource_filename("SiPANN", "LR/DC_coeffs.pkl")
DC_coeffs = ImportLR(cross_file)


#########################################################################################
######################  Helper Functions used throughout classes  #######################
#########################################################################################
def get_neff(wavelength, width, thickness, sw_angle=90):
    """Return neff for a given waveguide profile.

    Leverages Multivariate Linear Regression that maps wavelength, width, thickness and
    sidewall angle to effective index with silicon core and silicon dioxide cladding

    Parameters
    ----------
        wavelength:  float or ndarray
            wavelength (Valid for 1450nm-1650nm)
        width:    float or ndarray
            width (Valid for 400nm-600nm)
        thickness:    float or ndarray 
            thickness (Valid for 180nm-240nm)
        sw_angle:    float or ndarray
            sw_angle (Valid for 80-90 degrees)

    Returns
    ----------
        neff:    float or ndarray
            effective index of waveguide
    """

    # clean everything
    wavelength, width, thickness, sw_angle = clean_inputs(
        (wavelength, width, thickness, sw_angle)
    )
    # get coefficients
    _, _, _, _, neff = get_coeffs(wavelength, width, thickness, sw_angle)

    return neff


def get_coeffs(wavelength, width, thickness, sw_angle):
    """Return coefficients and neff for a given waveguide profile as used in
    SCEE.

    Leverages Multivariate Linear Regression that maps wavelength, width, thickness and
    sidewall angle to effective index and coefficients used in estimate of even and odd
    effective indices with silicon core and silicon dioxide cladding.

    Parameters
    ----------
        wavelength:    float or ndarray
            wavelength (Valid for 1450nm-1650nm)
        width:    float or ndarray
            width (Valid for 400nm-600nm)
        thickness:    float or ndarray
            thickness (Valid for 180nm-240nm)
        sw_angle:    float or ndarray
            sw_angle (Valid for 80-90 degrees)

    Returns
    ----------
        ae:    float or ndarray
            used in even mode estimation in neff + ae exp(ge * g)
        ao:    float or ndarray
            used in odd mode estimation in neff + ao exp(go * g)
        ge:    float or ndarray
            used in even mode estimation in neff + ae exp(ge * g)
        go:    float or ndarray
            used in odd mode estimation in neff + ao exp(go * g)
        neff:    float or ndarray
            effective index of waveguide
    """
    inputs = np.column_stack((wavelength, width, thickness, sw_angle))
    coeffs = DC_coeffs.predict(inputs)
    ae = coeffs[:, 0]
    ao = coeffs[:, 1]
    ge = coeffs[:, 2]
    go = coeffs[:, 3]
    neff = coeffs[:, 4]

    return (ae, ao, ge, go, neff)


def get_closed_ans(
    ae, ao, ge, go, neff, wavelength, gap, B, xe, xo, offset, trig, z_dist
):
    """Return coupling as found in Columbia paper.

    Uses general form of closed form solutions as found in M. Bahadori et al.,
    "Design Space Exploration of Microring Resonators in Silicon Photonic Interconnects: Impact of the Ring Curvature,"
    in Journal of Lightwave Technology, vol. 36, no. 13, pp. 2767-2782, 1 July1, 2018..

    Parameters
    ----------
        ae:    float or ndarray
            used in even mode estimation in neff + ae exp(ge * g)
        ao:    float or ndarray
            used in odd mode estimation in neff + ao exp(go * g)
        ge:    float or ndarray
            used in even mode estimation in neff + ae exp(ge * g)
        go:    float or ndarray
            used in odd mode estimation in neff + ao exp(go * g)
        neff:    float or ndarray
            effective index of waveguide
        wavelength:    float or ndarray (Valid for 1450nm-1650nm)
            wavelength
        gap:    float or ndarray
            gap distance
        B   (function): B function as found in paper
        xe:    float or ndarray
            as found in paper
        xo:    float or ndarray
            as found in paper
        offset:    float or ndarray
            0 or pi/2 depending on through/cross coupling
        trig:    float or ndarray
            sin or cos depending on through/cross coupling
        z_dist:    float or ndarray
            distance light will travel

    Returns
    ----------
        k/t   (complex ndarray): coupling coefficient
    """
    even_part = ae * np.exp(-ge * gap) * B(xe) / ge
    odd_part = ao * np.exp(-go * gap) * B(xo) / go
    phase_part = 2 * z_dist * neff

    mag = trig((even_part + odd_part) * np.pi / wavelength)
    phase = (even_part - odd_part + phase_part) * np.pi / wavelength + offset

    return mag * np.exp(-1j * phase)


def clean_inputs(inputs):
    """Makes all inputs as the same shape to allow passing arrays through.

    Used to make sure all inputs have the same length - ie that it's trying
    to run a specific number of simulations, not a varying amount

    Parameters
    ----------
        inputs : tuple
            can be an arbitrary mixture of floats/ndarray

    Returns
    ----------
        inputs : tuple
            returns all inputs as same size ndarrays
    """

    inputs = list(inputs)
    # make all scalars into numpy arrays
    for i in range(len(inputs)):
        if np.isscalar(inputs[i]):
            inputs[i] = np.array([inputs[i]])

    # take largest size of numpy arrays, or set value (if it's not 0)
    n = max(len(i) for i in inputs)

    # if it's smaller than largest, make array full of same value
    for i in range(len(inputs)):
        if len(inputs[i]) != n:
            if len(inputs[i]) != 1:
                raise ValueError("Mismatched Input Array Size")
            inputs[i] = np.full((n), inputs[i][0])

    return inputs


class DC(ABC):
    """Abstract Class that all directional couplers inherit from. Each DC will
    inherit from it.

    Base Class for DC. All other DC classes should be based on this one, including same functions (so
    documentation should be the same/similar with exception of device specific attributes). Ports are numbered as::

        |       2---\      /---4       |
        |            ------            |
        |            ------            |
        |       1---/      \---3       |

    Parameters
    ----------
        width : float or ndarray
            Width of the waveguide in nm (Valid for 400nm-600nm)
        thickness : float or ndarray
            Thickness of waveguide in nm (Valid for 180nm-240nm)
        sw_angle : float or ndarray, optional
            Sidewall angle of waveguide from horizontal in degrees (Valid for 80-90 degrees). Defaults to 90.
    """

    def __init__(self, width, thickness, sw_angle=90):
        self.width = width
        self.thickness = thickness
        self.sw_angle = sw_angle
        if np.any(self.width < 400):
            warnings.warn(
                "Width is less than 400nm, may produce invalid results", UserWarning
            )
        if np.any(self.width > 600):
            warnings.warn(
                "Width is greater than 600nm, may produce invalid results", UserWarning
            )
        if np.any(self.thickness < 180):
            warnings.warn(
                "Thickness is less than 180nm, may produce invalid results", UserWarning
            )
        if np.any(self.thickness > 240):
            warnings.warn(
                "Thickness is greater than 240nm, may produce invalid results",
                UserWarning,
            )
        if np.any(self.sw_angle < 80):
            warnings.warn(
                "Sidewall Angle is less than 80 degrees, may produce invalid results",
                UserWarning,
            )
        if np.any(self.sw_angle > 90):
            warnings.warn(
                "Sidewall Angle is greater than 90 degrees, may produce invalid results",
                UserWarning,
            )

    def _clean_args(self, wavelength):
        """Makes sure all attributes are the same size.

        Parses through all self attributes to make sure they're all the same size for
        simulations. Must be reimplemented for all child classes if they have unique attributes.
        Also takes in wavelength parameter to clean as is needed occasionally.

        Parameters
        ----------
        wavelength : float or ndarray
            Wavelength (Valid for 1450nm-1650nm)

        Returns
        ----------
        inputs : (tuple)
            Cleaned array of all devices attributes (and wavelength if included.)
        """
        if wavelength is None:
            return clean_inputs((self.width, self.thickness, self.sw_angle))
        else:
            return clean_inputs((wavelength, self.width, self.thickness, self.sw_angle))

    def update(self, **kwargs):
        """Takes in any parameter defined by __init__ and changes it.

        Parameters
        ----------
        attribute : float or ndarray
            Included if any device needs to have an attribute changed.
        """
        self.width = kwargs.get("width", self.width)
        self.thickness = kwargs.get("thickness", self.thickness)
        self.sw_angle = kwargs.get("sw_angle", self.sw_angle)
        if np.any(self.width < 400):
            warnings.warn(
                "Width is less than 400nm, may produce invalid results", UserWarning
            )
        if np.any(self.width > 600):
            warnings.warn(
                "Width is greater than 600nm, may produce invalid results", UserWarning
            )
        if np.any(self.thickness < 180):
            warnings.warn(
                "Thickness is less than 180nm, may produce invalid results", UserWarning
            )
        if np.any(self.thickness > 240):
            warnings.warn(
                "Thickness is greater than 240nm, may produce invalid results",
                UserWarning,
            )
        if np.any(self.sw_angle < 80):
            warnings.warn(
                "Sidewall Angle is less than 80 degrees, may produce invalid results",
                UserWarning,
            )
        if np.any(self.sw_angle > 90):
            warnings.warn(
                "Sidewall Angle is greater than 90 degrees, may produce invalid results",
                UserWarning,
            )

    def sparams(self, wavelength):
        """Returns scattering parameters.

        Runs SCEE to get scattering parameters at wavelength input.

        Parameters
        ----------
        wavelength:    float or ndarray
            wavelengths to get sparams at (Valid for 1450nm-1650nm)

        Returns
        ----------
        freq : ndarray
            frequency for s_matrix in Hz, size n (number of wavelength points)
        s_matrix : ndarray
            size (n,4,4) complex matrix of scattering parameters, in order of passed in wavelengths
        """
        # get number of points to evaluate at
        n = 1 if np.isscalar(wavelength) else len(wavelength)
        # check to make sure the geometry isn't an array
        if len(self._clean_args(None)[0]) != 1:
            raise ValueError(
                "You have changing geometries, getting sparams doesn't make sense"
            )
        s_matrix = np.zeros((4, 4, n), dtype="complex")

        # calculate upper half of matrix (diagonal is 0)
        for i in range(1, 5):
            for j in range(i, 5):
                s_matrix[i - 1, j - 1] = self.predict((i, j), wavelength)

        # apply symmetry (note diagonal is 0, no need to subtract it)
        s_matrix += np.transpose(s_matrix, (1, 0, 2))

        # transpose so depth comes first
        s_matrix = np.transpose(s_matrix, (2, 0, 1))
        return s_matrix

    @abstractmethod
    def predict(self, ports, wavelength):
        """Predicts the output when light is put in the specified port (see
        diagram above)

        Parameters
        ----------
        ports : 2-tuple
            Specifies the port coming in and coming out
        wavelength : float or ndarray
            Wavelength(s) to predict at (Valid for 1450nm-1650nm)
        extra_arc : float, optional
            Adds phase to compensate for waveguides getting to gap function. Defaults to 0.
        part : {"both", "mag", "ph"}, optional
            To speed up calculation, can calculate only magnitude (mag), phase (ph), or both. Defaults to both.

        Returns
        ----------
        k/t : complex ndarray
            The value of the light coming through
        """
        pass

    @abstractmethod
    def gds(
        self, filename=None, extra=0, units="microns", view=False, sbend_h=0, sbend_v=0
    ):
        """Writes the geometry to the gds file.

        Parameters
        ----------
        filename : str, optional
            Location to save file to. Defaults to None.
        extra : int, optional
            Extra straight portion to add to ends of waveguides to make room in simulation
            (units same as units parameter). Defaults to 0.
        units : {'microns' or 'nms'}, optional
            Units to save gds file in. Defaults to microns.
        view : bool, optional
            Whether to visually show gds file. Defaults to False.
        sbend_h : int, optional
            How high to horizontally make additional sbends to move ports farther away.
            Sbends insert after extra. Only available in couplers with all horizontal
            ports (units same as units parameters). Defaults to 0
        sbend_v : int, optional
            Same as sbend_h, but vertical distance. Defaults to 0.
        """
        pass


#########################################################################################
# Integral Estimators. Make any coupling device as desired.
#########################################################################################
class GapFuncSymmetric(DC):
    """This class will create arbitrarily shaped SYMMETRIC (ie both waveguides
    are same shape) directional couplers.

    It takes in a gap function that describes the gap as one progreses through the device. Note that the shape fo the waveguide
    will simply be half of gap function. Also requires the derivative of the gap function for this purpose. Ports are numbered as::

        |       2---\      /---4       |
        |            ------            |
        |            ------            |
        |       1---/      \---3       |

    Parameters
    ----------
        width : float or ndarray
            Width of the waveguide in nm (Valid for 400nm-600nm)
        thickness : float or ndarray
            Thickness of waveguide in nm (Valid for 180nm-240nm)
        gap : function
            Gap function as one progresses along the waveguide (Must always be > 100nm)
        dgap : function
            Derivative of the gap function
        zmin : float
            Where to begin integration in the gap function
        zmax : float
            Where to end integration in the gap function
        sw_angle : float or ndarray, optional
            Sidewall angle of waveguide from horizontal in degrees (Valid for 80-90 degrees). Defaults to 90.
    """

    def __init__(self, width, thickness, gap, dgap, zmin, zmax, sw_angle=90):
        super().__init__(width, thickness, sw_angle)
        self.gap = gap
        self.dgap = dgap
        self.zmin = zmin
        self.zmax = zmax

    def update(self, **kwargs):
        super().update(**kwargs)
        self.gap = kwargs.get("gap", self.gap)
        self.dgap = kwargs.get("dgap", self.dgap)
        self.zmin = kwargs.get("zmin", self.zmin)
        self.zmax = kwargs.get("zmax", self.zmax)

    def _clean_args(self, wavelength):
        if wavelength is None:
            return clean_inputs((self.width, self.thickness, self.sw_angle))
        else:
            return clean_inputs((wavelength, self.width, self.thickness, self.sw_angle))

    def predict(self, ports, wavelength, extra_arc=0, part="both"):
        # check to make sure wavelength is in valid range
        if np.any(wavelength < 1450):
            warnings.warn(
                "Wavelength is less than 1450nm, may produce invalid results",
                UserWarning,
            )
        if np.any(wavelength > 1650):
            warnings.warn(
                "Wavelength is greater than 1650nm, may produce invalid results",
                UserWarning,
            )

        # clean data and get coefficients
        wavelength, width, thickness, sw_angle = self._clean_args(wavelength)
        n = len(wavelength)
        ae, ao, ge, go, neff = get_coeffs(wavelength, width, thickness, sw_angle)

        # make sure ports are valid
        if not all(1 <= x <= 4 for x in ports):
            raise ValueError("Invalid Ports")

        # if it's coming to itself, or to adjacent port
        if (
            (ports[0] == ports[1])
            or (ports[0] + ports[1] == 3)
            or (ports[0] + ports[1] == 7)
        ):
            return np.zeros(len(wavelength))

        # determine if cross or through port
        if abs(ports[1] - ports[0]) == 2:
            trig = np.cos
            offset = 0
        else:
            trig = np.sin
            offset = np.pi / 2

        # determine z distance
        arcFomula = lambda x: np.sqrt(1 + (self.dgap(x) / 2) ** 2)
        z_dist = quad(arcFomula, self.zmin, self.zmax)[0] + extra_arc

        # calculate everything
        mag = np.ones(n)
        phase = np.zeros(n)
        for i in range(n):
            if part in ["both", "mag"]:
                f_mag = lambda z: float(
                    ae[i] * np.exp(-ge[i] * self.gap(z))
                    + ao[i] * np.exp(-go[i] * self.gap(z))
                )
                mag[i] = trig(
                    np.pi * quad(f_mag, self.zmin, self.zmax)[0] / wavelength[i]
                )
            if part in ["both", "ph"]:
                f_phase = lambda z: float(
                    ae[i] * np.exp(-ge[i] * self.gap(z))
                    - ao[i] * np.exp(-go[i] * self.gap(z))
                )
                phase[i] = (
                    np.pi * quad(f_phase, self.zmin, self.zmax)[0] / wavelength[i]
                    + 2 * np.pi * neff[i] * z_dist / wavelength[i]
                    + offset
                )

        return mag * np.exp(-1j * phase)

    def gds(
        self, filename=None, extra=0, units="microns", view=False, sbend_h=0, sbend_v=0
    ):
        # check to make sure the geometry isn't an array
        if len(self._clean_args(None)[0]) != 1:
            raise ValueError(
                "You have changing geometries, making gds doesn't make sense"
            )

        if units == "nms":
            scale = 1
        elif units == "microns":
            scale = 10 ** -3
        else:
            raise ValueError("Invalid units")

        # scale to proper units
        sc_zmin = self.zmin * scale
        sc_zmax = self.zmax * scale
        sc_width = self.width * scale
        cL = sc_zmax - sc_zmin
        cH = self.gap(self.zmin) * scale / 2

        # make parametric functions
        paraTop = lambda x: (
            x * cL + sc_zmin,
            scale * self.gap(x * (self.zmax - self.zmin) + self.zmin) / 2
            + sc_width / 2,
        )
        paraBottom = lambda x: (
            x * cL + sc_zmin,
            -scale * self.gap(x * (self.zmax - self.zmin) + self.zmin) / 2
            - sc_width / 2,
        )
        sbend = sbend_h != 0 and sbend_v != 0
        sbendDown = lambda x: (sbend_h * x, -sbend_v / 2 * (1 - np.cos(np.pi * x)))
        sbendUp = lambda x: (sbend_h * x, sbend_v / 2 * (1 - np.cos(np.pi * x)))
        dsbendDown = lambda x: (sbend_h, -np.pi * sbend_v / 2 * np.sin(np.pi * x))
        dsbendUp = lambda x: (sbend_h, np.pi * sbend_v / 2 * np.sin(np.pi * x))

        # write to GDS
        pathTop = gdspy.Path(
            sc_width, (sc_zmin - extra - sbend_h, cH + sc_width / 2 + sbend_v)
        )
        pathTop.segment(extra, "+x")
        if sbend:
            pathTop.parametric(sbendDown, dsbendDown)
        pathTop.parametric(paraTop, relative=False)
        if sbend:
            pathTop.parametric(sbendUp, dsbendUp)
        pathTop.segment(extra, "+x")

        pathBottom = gdspy.Path(
            sc_width, (sc_zmin - extra - sbend_h, -cH - sc_width / 2 - sbend_v)
        )
        pathBottom.segment(extra, "+x")
        if sbend:
            pathBottom.parametric(sbendUp, dsbendUp)
        pathBottom.parametric(paraBottom, relative=False)
        if sbend:
            pathBottom.parametric(sbendDown, dsbendDown)
        pathBottom.segment(extra, "+x")

        gdspy.current_library = gdspy.GdsLibrary()
        path_cell = gdspy.Cell("C0")
        path_cell.add(pathTop)
        path_cell.add(pathBottom)

        if view:
            gdspy.LayoutViewer(cells=path_cell)

        if filename is not None:
            writer = gdspy.GdsWriter(filename, unit=1.0e-6, precision=1.0e-9)
            writer.write_cell(path_cell)
            writer.close()


class GapFuncAntiSymmetric(DC):
    """This class will create arbitrarily shaped ANTISYMMETRIC (ie waveguides
    are different shapes) directional couplers.

    It takes in a gap function that describes the gap as one progreses through the device. Also takes in arc length
    of each port up till coupling point.
    Ports are numbered as:
    |       2---\      /---4       |
    |            ------            |
    |            ------            |
    |       1---/      \---3       |

    Parameters
    ----------
        width : float or ndarray
            Width of the waveguide in nm (Valid for 400nm-600nm)
        thickness : float or ndarray
            Thickness of waveguide in nm (Valid for 180nm-240nm)
        gap : function
            Gap function as one progresses along the waveguide (Must always be > 100nm)
        zmin : float
            Where to begin integration in the gap function
        zmax : float
            Where to end integration in the gap function
        arc1, arc2, arc3, arc4 : float
            Arclength from entrance of each port till minimum coupling point
        sw_angle : float or ndarray, optional
            Sidewall angle of waveguide from horizontal in degrees (Valid for 80-90 degrees). Defaults to 90.
    """

    def __init__(
        self, width, thickness, gap, zmin, zmax, arc1, arc2, arc3, arc4, sw_angle=90
    ):
        super().__init__(width, thickness, sw_angle)
        self.gap = gap
        self.zmin = zmin
        self.zmax = zmax
        self.arc1 = arc1
        self.arc2 = arc2
        self.arc3 = arc3
        self.arc4 = arc4

    def update(self, **kwargs):
        super().update(**kwargs)
        self.gap = kwargs.get("gap", self.gap)
        self.arc_l = kwargs.get("arc_l", self.arc_l)
        self.arc_u = kwargs.get("arc_u", self.arc_u)
        self.zmin = kwargs.get("zmin", self.zmin)
        self.zmax = kwargs.get("zmax", self.zmax)

    def _clean_args(self, wavelength):
        if wavelength is None:
            return clean_inputs((self.width, self.thickness, self.sw_angle))
        else:
            return clean_inputs((wavelength, self.width, self.thickness, self.sw_angle))

    def predict(self, ports, wavelength, extra_arc=0, part="both"):
        # check to make sure wavelength is in valid range
        if np.any(wavelength < 1450):
            warnings.warn(
                "Wavelength is less than 1450nm, may produce invalid results",
                UserWarning,
            )
        if np.any(wavelength > 1650):
            warnings.warn(
                "Wavelength is greater than 1650nm, may produce invalid results",
                UserWarning,
            )

        # clean data and get coefficients
        wavelength, width, thickness, sw_angle = self._clean_args(wavelength)
        n = len(wavelength)
        ae, ao, ge, go, neff = get_coeffs(wavelength, width, thickness, sw_angle)

        # make sure ports are valid
        if not all(1 <= x <= 4 for x in ports):
            raise ValueError("Invalid Ports")

        # determine if cross or through port
        if abs(ports[1] - ports[0]) == 2:
            trig = np.cos
            offset = 0
        else:
            trig = np.sin
            offset = np.pi / 2

        # determine z distance
        if 1 in ports and 3 in ports:
            z_dist = self.arc1 + self.arc3 + extra_arc
        elif 1 in ports and 4 in ports:
            z_dist = self.arc1 + self.arc4 + extra_arc
        elif 2 in ports and 4 in ports:
            z_dist = self.arc2 + self.arc4 + extra_arc
        elif 2 in ports and 3 in ports:
            z_dist = self.arc2 + self.arc3 + extra_arc
        # if it's coming to itself, or to adjacent port
        else:
            return np.zeros(len(wavelength))

        # calculate everything
        mag = np.ones(n)
        phase = np.zeros(n)
        for i in range(n):
            if part in ["both", "mag"]:
                f_mag = lambda z: float(
                    ae[i] * np.exp(-ge[i] * self.gap(z))
                    + ao[i] * np.exp(-go[i] * self.gap(z))
                )
                mag[i] = trig(
                    np.pi * quad(f_mag, self.zmin, self.zmax)[0] / wavelength[i]
                )
            if part in ["both", "ph"]:
                f_phase = lambda z: float(
                    ae[i] * np.exp(-ge[i] * self.gap(z))
                    - ao[i] * np.exp(-go[i] * self.gap(z))
                )
                phase[i] = (
                    np.pi * quad(f_phase, self.zmin, self.zmax)[0] / wavelength[i]
                    + 2 * np.pi * neff[i] * z_dist / wavelength[i]
                    + offset
                )

        return mag * np.exp(-1j * phase)

    def gds(
        self, filename=None, extra=0, units="microns", view=False, sbend_h=0, sbend_v=0
    ):
        """Still needs to be implemented for this class."""
        raise NotImplementedError(
            "Generating GDS file of nonsymmetric coupler not supported yet."
        )


#########################################################################################
# All the Different types of DC's with closed form solutions. These will be faster than defining it manually in the function form above.
#########################################################################################
class HalfRing(DC):
    """This class will create half of a ring resonator.

    It takes in a radius and gap along with usual waveguide parameters. Ports are numbered as::

        |         2 \     / 4          |
        |            \   /             |
        |             ---              |
        |         1---------3          |

    Parameters
    ----------
        width : float or ndarray
            Width of the waveguide in nm (Valid for 400nm-600nm)
        thickness : float or ndarray
            Thickness of waveguide in nm (Valid for 180nm-240nm)
        radius : float or ndarray
            Distance from center of ring to middle of waveguide in nm.
        gap : float or ndarray
            Minimum distance from ring waveguide edge to straight waveguide edge in nm. (Must be > 100nm)
        sw_angle : float or ndarray, optional
            Sidewall angle of waveguide from horizontal in degrees (Valid for 80-90 degrees). Defaults to 90.
    """

    def __init__(self, width, thickness, radius, gap, sw_angle=90):
        super().__init__(width, thickness, sw_angle)
        self.radius = radius
        self.gap = gap
        if np.any(self.gap < 100):
            warnings.warn(
                "Gap is less than 100nm, may produce invalid results", UserWarning
            )

    def update(self, **kwargs):
        super().update(**kwargs)
        self.radius = kwargs.get("radius", self.radius)
        self.gap = kwargs.get("gap", self.gap)
        if np.any(self.gap < 100):
            warnings.warn(
                "Gap is less than 100nm, may produce invalid results", UserWarning
            )

    def _clean_args(self, wavelength):
        if wavelength is None:
            return clean_inputs(
                (self.width, self.thickness, self.sw_angle, self.radius, self.gap)
            )
        else:
            return clean_inputs(
                (
                    wavelength,
                    self.width,
                    self.thickness,
                    self.sw_angle,
                    self.radius,
                    self.gap,
                )
            )

    def predict(self, ports, wavelength):
        # check to make sure wavelength is in valid range
        if np.any(wavelength < 1450):
            warnings.warn(
                "Wavelength is less than 1450nm, may produce invalid results",
                UserWarning,
            )
        if np.any(wavelength > 1650):
            warnings.warn(
                "Wavelength is greater than 1650nm, may produce invalid results",
                UserWarning,
            )

        # clean data and get coefficients
        wavelength, width, thickness, sw_angle, radius, gap = self._clean_args(
            wavelength
        )
        ae, ao, ge, go, neff = get_coeffs(wavelength, width, thickness, sw_angle)

        # make sure ports are valid
        if not all(1 <= x <= 4 for x in ports):
            raise ValueError("Invalid Ports")

        # determine if cross or through port
        if abs(ports[1] - ports[0]) == 2:
            trig = np.cos
            offset = 0
        else:
            trig = np.sin
            offset = np.pi / 2

        # determine z distance
        if 1 in ports and 3 in ports:
            z_dist = 2 * (radius + width / 2)
        elif 1 in ports and 4 in ports or (2 not in ports or 4 not in ports) and 2 in ports and 3 in ports:
            z_dist = np.pi * radius / 2 + radius + width / 2
        elif 2 in ports and 4 in ports:
            z_dist = np.pi * radius
        else:
            return np.zeros(len(wavelength))

        # calculate everything
        B = (
            lambda x: np.pi
            * x
            * np.exp(-x)
            * (special.iv(1, x) + special.modstruve(-1, x))
        )
        xe = ge * (radius + width / 2)
        xo = go * (radius + width / 2)
        return get_closed_ans(
            ae, ao, ge, go, neff, wavelength, gap, B, xe, xo, offset, trig, z_dist
        )

    def gds(self, filename=None, view=False, extra=0, units="nms"):
        """Writes the geometry to the gds file.

        Parameters
        ----------
        filename : str, optional
            Location to save file to. Defaults to None.
        extra : int, optional
            Extra straight portion to add to ends of waveguides to make room in simulation
            (units same as units parameter). Defaults to 0.
        units : {'microns' or 'nms'}, optional
            Units to save gds file in. Defaults to microns.
        view : bool, optional
            Whether to visually show gds file. Defaults to False.
        """
        # check to make sure the geometry isn't an array
        if len(self._clean_args(None)[0]) != 1:
            raise ValueError(
                "You have changing geometries, making gds doesn't make sense"
            )

        if units == "nms":
            scale = 1
        elif units == "microns":
            scale = 10 ** -3
        else:
            raise ValueError("Invalid units")

        # scale to proper units
        sc_radius = self.radius * scale
        sc_gap = self.gap * scale
        sc_width = self.width * scale

        # write to GDS
        pathTop = gdspy.Path(
            sc_width, (sc_radius, sc_radius + sc_width / 2 + sc_gap / 2 + extra)
        )
        pathTop.segment(extra, "-y")
        pathTop.arc(sc_radius, 0, -np.pi)
        pathTop.segment(extra, "+y")

        pathBottom = gdspy.Path(
            sc_width, (-sc_radius - sc_width / 2 - extra, -sc_gap / 2 - sc_width / 2)
        )
        pathBottom.segment(2 * (sc_radius + sc_width / 2 + extra), "+x")

        gdspy.current_library = gdspy.GdsLibrary()
        path_cell = gdspy.Cell("C0")
        path_cell.add(pathTop)
        path_cell.add(pathBottom)

        if view:
            gdspy.LayoutViewer(cells=path_cell)

        if filename is not None:
            writer = gdspy.GdsWriter(filename, unit=1.0e-6, precision=1.0e-9)
            writer.write_cell(path_cell)
            writer.close()


class HalfRacetrack(DC):
    """This class will create half of a ring resonator.

    It takes in a radius and gap along with usual waveguide parameters. Ports are numbered as::

        |      2 \           / 4       |
        |         \         /          |
        |          ---------           |
        |      1---------------3       |

    Parameters
    ----------
        width : float or ndarray
            Width of the waveguide in nm (Valid for 400nm-600nm)
        thickness : float or ndarray
            Thickness of waveguide in nm (Valid for 180nm-240nm)
        radius : float or ndarray
            Distance from center of ring to middle of waveguide in nm.
        gap : float or ndarray
            Minimum distance from ring waveguide edge to straight waveguide edge in nm. (Must be > 100nm)
        length : float or ndarray
            Length of straight portion of ring waveguide in nm.
        sw_angle : float or ndarray, optional
            Sidewall angle of waveguide from horizontal in degrees (Valid for 80-90 degrees). Defaults to 90.
    """

    def __init__(self, width, thickness, radius, gap, length, sw_angle=90):
        super().__init__(width, thickness, sw_angle)
        self.radius = radius
        self.gap = gap
        self.length = length
        if np.any(self.gap < 100):
            warnings.warn(
                "Gap is less than 100nm, may produce invalid results", UserWarning
            )

    def update(self, **kwargs):
        super().update(**kwargs)
        self.radius = kwargs.get("radius", self.radius)
        self.gap = kwargs.get("gap", self.gap)
        self.length = kwargs.get("length", self.length)
        if np.any(self.gap < 100):
            warnings.warn(
                "Gap is less than 100nm, may produce invalid results", UserWarning
            )

    def _clean_args(self, wavelength):
        if wavelength is None:
            return clean_inputs(
                (
                    self.width,
                    self.thickness,
                    self.sw_angle,
                    self.radius,
                    self.gap,
                    self.length,
                )
            )
        else:
            return clean_inputs(
                (
                    wavelength,
                    self.width,
                    self.thickness,
                    self.sw_angle,
                    self.radius,
                    self.gap,
                    self.length,
                )
            )

    def predict(self, ports, wavelength):
        # check to make sure wavelength is in valid range
        if np.any(wavelength < 1450):
            warnings.warn(
                "Wavelength is less than 1450nm, may produce invalid results",
                UserWarning,
            )
        if np.any(wavelength > 1650):
            warnings.warn(
                "Wavelength is greater than 1650nm, may produce invalid results",
                UserWarning,
            )

        # clean data and get coefficients
        wavelength, width, thickness, sw_angle, radius, gap, length = self._clean_args(
            wavelength
        )
        ae, ao, ge, go, neff = get_coeffs(wavelength, width, thickness, sw_angle)

        # make sure ports are valid
        if not all(1 <= x <= 4 for x in ports):
            raise ValueError("Invalid Ports")

        # determine if cross or through port
        if abs(ports[1] - ports[0]) == 2:
            trig = np.cos
            offset = 0
        else:
            trig = np.sin
            offset = np.pi / 2

        # determine z distance
        if 1 in ports and 3 in ports:
            z_dist = 2 * (radius + width / 2) + length
        elif 1 in ports and 4 in ports or (2 not in ports or 4 not in ports) and 2 in ports and 3 in ports:
            z_dist = np.pi * radius / 2 + radius + width / 2 + length
        elif 2 in ports and 4 in ports:
            z_dist = np.pi * radius + length
        else:
            return np.zeros(len(wavelength))

        # calculate everything
        B = lambda x: length * x / (radius + width / 2) + np.pi * x * np.exp(-x) * (
            special.iv(1, x) + special.modstruve(-1, x)
        )
        xe = ge * (radius + width / 2)
        xo = go * (radius + width / 2)
        return get_closed_ans(
            ae, ao, ge, go, neff, wavelength, gap, B, xe, xo, offset, trig, z_dist
        )

    def gds(self, filename=None, view=False, extra=0, units="nms"):
        """Writes the geometry to the gds file.

        Parameters
        ----------
        filename : str, optional
            Location to save file to. Defaults to None.
        extra : int, optional
            Extra straight portion to add to ends of waveguides to make room in simulation
            (units same as units parameter). Defaults to 0.
        units : {'microns' or 'nms'}, optional
            Units to save gds file in. Defaults to microns.
        view : bool, optional
            Whether to visually show gds file. Defaults to False.
        """
        # check to make sure the geometry isn't an array
        if len(self._clean_args(None)[0]) != 1:
            raise ValueError(
                "You have changing geometries, making gds doesn't make sense"
            )

        if units == "nms":
            scale = 1
        elif units == "microns":
            scale = 10 ** -3
        else:
            raise ValueError("Invalid units")

        # scale to proper units
        sc_radius = self.radius * scale
        sc_gap = self.gap * scale
        sc_width = self.width * scale
        sc_length = self.length * scale

        # write to GDS
        pathTop = gdspy.Path(
            sc_width,
            (sc_radius + sc_length / 2, sc_radius + sc_width / 2 + sc_gap / 2 + extra),
        )
        pathTop.segment(extra, "-y")
        pathTop.arc(sc_radius, 0, -np.pi / 2)
        pathTop.segment(sc_length, "-x")
        pathTop.arc(sc_radius, -np.pi / 2, -np.pi)
        pathTop.segment(extra, "+y")

        pathBottom = gdspy.Path(
            sc_width,
            (
                -sc_radius - sc_width / 2 - sc_length / 2 - extra,
                -sc_gap / 2 - sc_width / 2,
            ),
        )
        pathBottom.segment(2 * (sc_radius + sc_width / 2) + sc_length + 2 * extra, "+x")

        gdspy.current_library = gdspy.GdsLibrary()
        path_cell = gdspy.Cell("C0")
        path_cell.add(pathTop)
        path_cell.add(pathBottom)

        if view:
            gdspy.LayoutViewer(cells=path_cell)

        if filename is not None:
            writer = gdspy.GdsWriter(filename, unit=1.0e-6, precision=1.0e-9)
            writer.write_cell(path_cell)
            writer.close()


class StraightCoupler(DC):
    """This class will create half of a ring resonator.

    It takes in a radius and gap along with usual waveguide parameters. Ports are numbered as::

        |      2---------------4       |
        |      1---------------3       |

    Parameters
    ----------
        width : float or ndarray
            Width of the waveguide in nm (Valid for 400nm-600nm)
        thickness : float or ndarray
            Thickness of waveguide in nm (Valid for 180nm-240nm)
        gap : float or ndarray
           Distance between the two waveguides edge in nm. (Must be > 100nm)
        length : float or ndarray
            Length of both waveguides in nm.
        sw_angle : float or ndarray, optional
            Sidewall angle of waveguide from horizontal in degrees (Valid for 80-90 degrees). Defaults to 90.
    """

    def __init__(self, width, thickness, gap, length, sw_angle=90):
        super().__init__(width, thickness, sw_angle)
        self.gap = gap
        self.length = length
        if np.any(self.gap < 100):
            warnings.warn(
                "Gap is less than 100nm, may produce invalid results", UserWarning
            )

    def update(self, **kwargs):
        super().update(**kwargs)
        self.gap = kwargs.get("gap", self.gap)
        self.length = kwargs.get("length", self.length)
        if np.any(self.gap < 100):
            warnings.warn(
                "Gap is less than 100nm, may produce invalid results", UserWarning
            )

    def _clean_args(self, wavelength):
        if wavelength is None:
            return clean_inputs(
                (self.width, self.thickness, self.sw_angle, self.gap, self.length)
            )
        else:
            return clean_inputs(
                (
                    wavelength,
                    self.width,
                    self.thickness,
                    self.sw_angle,
                    self.gap,
                    self.length,
                )
            )

    def predict(self, ports, wavelength):
        # check to make sure wavelength is in valid range
        if np.any(wavelength < 1450):
            warnings.warn(
                "Wavelength is less than 1450nm, may produce invalid results",
                UserWarning,
            )
        if np.any(wavelength > 1650):
            warnings.warn(
                "Wavelength is greater than 1650nm, may produce invalid results",
                UserWarning,
            )

        # clean data and get coefficients
        wavelength, width, thickness, sw_angle, gap, length = self._clean_args(
            wavelength
        )
        ae, ao, ge, go, neff = get_coeffs(wavelength, width, thickness, sw_angle)

        # make sure ports are valid
        if not all(1 <= x <= 4 for x in ports):
            raise ValueError("Invalid Ports")

        # determine if cross or through port
        if abs(ports[1] - ports[0]) == 2:
            trig = np.cos
            offset = 0
        else:
            trig = np.sin
            offset = np.pi / 2

        # determine z distance
        if 1 in ports and 3 in ports or 1 in ports and 4 in ports or 2 in ports and 4 in ports or 2 in ports and 3 in ports:
            z_dist = length
        else:
            return np.zeros(len(wavelength))

        # calculate everything
        B = lambda x: x
        xe = ge * length
        xo = go * length
        return get_closed_ans(
            ae, ao, ge, go, neff, wavelength, gap, B, xe, xo, offset, trig, z_dist
        )

    def gds(
        self, filename=None, view=False, extra=0, units="nms", sbend_h=0, sbend_v=0
    ):
        # check to make sure the geometry isn't an array
        if len(self._clean_args(None)[0]) != 1:
            raise ValueError(
                "You have changing geometries, making gds doesn't make sense"
            )

        if units == "nms":
            scale = 1
        elif units == "microns":
            scale = 10 ** -3
        else:
            raise ValueError("Invalid units")

        # scale to proper units
        sc_width = self.width * scale
        sc_gap = self.gap * scale
        sc_length = self.length * scale

        sbend = sbend_h != 0 and sbend_v != 0
        sbendDown = lambda x: (sbend_h * x, -sbend_v / 2 * (1 - np.cos(np.pi * x)))
        sbendUp = lambda x: (sbend_h * x, sbend_v / 2 * (1 - np.cos(np.pi * x)))
        dsbendDown = lambda x: (sbend_h, -np.pi * sbend_v / 2 * np.sin(np.pi * x))
        dsbendUp = lambda x: (sbend_h, np.pi * sbend_v / 2 * np.sin(np.pi * x))

        # write to GDS
        pathTop = gdspy.Path(
            sc_width,
            (-sc_length / 2 - sbend_h - extra, sbend_v + sc_width / 2 + sc_gap / 2),
        )
        pathTop.segment(extra, "+x")
        if sbend:
            pathTop.parametric(sbendDown, dsbendDown)
        pathTop.segment(sc_length, "+x")
        if sbend:
            pathTop.parametric(sbendUp, dsbendUp)
        pathTop.segment(extra, "+x")

        pathBottom = gdspy.Path(
            sc_width,
            (-sc_length / 2 - sbend_h - extra, -sbend_v - sc_width / 2 - sc_gap / 2),
        )
        pathBottom.segment(extra, "+x")
        if sbend:
            pathBottom.parametric(sbendUp, dsbendUp)
        pathBottom.segment(sc_length, "+x")
        if sbend:
            pathBottom.parametric(sbendDown, dsbendDown)
        pathBottom.segment(extra, "+x")

        gdspy.current_library = gdspy.GdsLibrary()
        path_cell = gdspy.Cell("C0")
        path_cell.add(pathTop)
        path_cell.add(pathBottom)

        if view:
            gdspy.LayoutViewer(cells=path_cell)

        if filename is not None:
            writer = gdspy.GdsWriter(filename, unit=1.0e-6, precision=1.0e-9)
            writer.write_cell(path_cell)
            writer.close()


class Standard(DC):
    """Normal/Standard Shaped Directional Coupler.

    This is what most people think of when they think directional coupler. Ports are numbered as::

        |       2---\      /---4       |
        |            ------            |
        |            ------            |
        |       1---/      \---3       |

    Parameters
    ----------
        width : float or ndarray
            Width of the waveguide in nm (Valid for 400nm-600nm)
        thickness : float or ndarray
            Thickness of waveguide in nm (Valid for 180nm-240nm)
        gap : float or ndarray
           Minimum distance between the two waveguides edge in nm. (Must be > 100nm)
        length : float or ndarray
            Length of the straight portion of both waveguides in nm.
        H : float or ndarray
            Horizontal distance between end of coupler until straight portion in nm.
        H : float or ndarray
            Vertical distance between end of coupler until straight portion in nm.
        sw_angle : float or ndarray, optional
            Sidewall angle of waveguide from horizontal in degrees (Valid for 80-90 degrees). Defaults to 90.
    """

    def __init__(self, width, thickness, gap, length, H, V, sw_angle=90):
        super().__init__(width, thickness, sw_angle)
        self.gap = gap
        self.length = length
        self.H = H
        self.V = V
        if np.any(self.gap < 100):
            warnings.warn(
                "Gap is less than 100nm, may produce invalid results", UserWarning
            )

    def update(self, **kwargs):
        super().update(**kwargs)
        self.gap = kwargs.get("gap", self.gap)
        self.length = kwargs.get("length", self.length)
        self.H = kwargs.get("H", self.H)
        self.V = kwargs.get("V", self.V)
        if np.any(self.gap < 100):
            warnings.warn(
                "Gap is less than 100nm, may produce invalid results", UserWarning
            )

    def _clean_args(self, wavelength):
        if wavelength is None:
            return clean_inputs(
                (
                    self.width,
                    self.thickness,
                    self.sw_angle,
                    self.gap,
                    self.length,
                    self.H,
                    self.V,
                )
            )
        else:
            return clean_inputs(
                (
                    wavelength,
                    self.width,
                    self.thickness,
                    self.sw_angle,
                    self.gap,
                    self.length,
                    self.H,
                    self.V,
                )
            )

    def predict(self, ports, wavelength):
        # check to make sure wavelength is in valid range
        if np.any(wavelength < 1450):
            warnings.warn(
                "Wavelength is less than 1450nm, may produce invalid results",
                UserWarning,
            )
        if np.any(wavelength > 1650):
            warnings.warn(
                "Wavelength is greater than 1650nm, may produce invalid results",
                UserWarning,
            )

        # clean data and get coefficients
        wavelength, width, thickness, sw_angle, gap, length, H, V = self._clean_args(
            wavelength
        )
        ae, ao, ge, go, neff = get_coeffs(wavelength, width, thickness, sw_angle)

        # make sure ports are valid
        if not all(1 <= x <= 4 for x in ports):
            raise ValueError("Invalid Ports")

        # determine if cross or through port
        if abs(ports[1] - ports[0]) == 2:
            trig = np.cos
            offset = 0
        else:
            trig = np.sin
            offset = np.pi / 2

        # determine z distance - length + 2*sbend length
        m = (V * np.pi / 2) ** 2 / (H ** 2 + (V * np.pi / 2) ** 2)
        z_dist = length + 2 * np.sqrt(
            H ** 2 + (V * np.pi / 2) ** 2
        ) / np.pi * special.ellipeinc(np.pi, m)
        if 1 in ports and 3 in ports or 1 in ports and 4 in ports or 2 in ports and 4 in ports or 2 in ports and 3 in ports:
            z_dist = z_dist
        else:
            return np.zeros(len(wavelength))

        # calculate everything
        B = lambda x: x * (
            1 + 2 * H * np.exp(-V * x / length) * special.iv(0, V * x / length) / length
        )
        xe = ge * length
        xo = go * length
        return get_closed_ans(
            ae, ao, ge, go, neff, wavelength, gap, B, xe, xo, offset, trig, z_dist
        )

    def gds(
        self, filename=None, view=False, extra=0, units="nms", sbend_h=0, sbend_v=0
    ):
        # check to make sure the geometry isn't an array
        if len(self._clean_args(None)[0]) != 1:
            raise ValueError(
                "You have changing geometries, making gds doesn't make sense"
            )

        if units == "nms":
            scale = 1
        elif units == "microns":
            scale = 10 ** -3
        else:
            raise ValueError("Invalid units")

        # scale to proper units
        sc_width = self.width * scale
        sc_gap = self.gap * scale
        sc_length = self.length * scale
        sc_H = self.H * scale
        sc_V = self.V * scale

        # make parametric functions
        sbendDown = lambda x: (sc_H * x, -sc_V / 2 * (1 - np.cos(np.pi * x)))
        sbendUp = lambda x: (sc_H * x, sc_V / 2 * (1 - np.cos(np.pi * x)))
        dsbendDown = lambda x: (sc_H, -np.pi * sc_V / 2 * np.sin(np.pi * x))
        dsbendUp = lambda x: (sc_H, np.pi * sc_V / 2 * np.sin(np.pi * x))

        sbend = sbend_h != 0 and sbend_v != 0
        sbendDownExtra = lambda x: (sbend_h * x, -sbend_v / 2 * (1 - np.cos(np.pi * x)))
        sbendUpExtra = lambda x: (sbend_h * x, sbend_v / 2 * (1 - np.cos(np.pi * x)))
        dsbendDownExtra = lambda x: (sbend_h, -np.pi * sbend_v / 2 * np.sin(np.pi * x))
        dsbendUpExtra = lambda x: (sbend_h, np.pi * sbend_v / 2 * np.sin(np.pi * x))

        # write to GDS
        pathTop = gdspy.Path(
            sc_width,
            (
                -sc_length / 2 - sc_H - sbend_h - extra,
                sc_V + sbend_v + sc_width / 2 + sc_gap / 2,
            ),
        )
        pathTop.segment(extra, "+x")
        if sbend:
            pathTop.parametric(sbendDownExtra, dsbendDownExtra)
        pathTop.parametric(sbendDown, dsbendDown)
        pathTop.segment(sc_length, "+x")
        pathTop.parametric(sbendUp, dsbendUp)
        if sbend:
            pathTop.parametric(sbendUpExtra, dsbendUpExtra)
        pathTop.segment(extra, "+x")

        pathBottom = gdspy.Path(
            sc_width,
            (
                -sc_length / 2 - sc_H - sbend_h - extra,
                -sc_V - sbend_v - sc_width / 2 - sc_gap / 2,
            ),
        )
        pathBottom.segment(extra, "+x")
        if sbend:
            pathBottom.parametric(sbendUpExtra, dsbendUpExtra)
        pathBottom.parametric(sbendUp, dsbendUp)
        pathBottom.segment(sc_length, "+x")
        pathBottom.parametric(sbendDown, dsbendDown)
        if sbend:
            pathBottom.parametric(sbendDownExtra, dsbendDownExtra)
        pathBottom.segment(extra, "+x")

        gdspy.current_library = gdspy.GdsLibrary()
        path_cell = gdspy.Cell("C0")
        path_cell.add(pathTop)
        path_cell.add(pathBottom)

        if view:
            gdspy.LayoutViewer(cells=path_cell)

        if filename is not None:
            writer = gdspy.GdsWriter(filename, unit=1.0e-6, precision=1.0e-9)
            writer.write_cell(path_cell)
            writer.close()


class DoubleHalfRing(DC):
    """This class will create two equally sized halfrings coupling.

    It takes in a radius and gap along with usual waveguide parameters. Ports are numbered as::

        |         2 \     / 4          |
        |            \   /             |
        |             ---              |
        |             ---              |
        |            /   \             |
        |         1 /     \ 3          |

    Parameters
    ----------
        width : float or ndarray
            Width of the waveguide in nm (Valid for 400nm-600nm)
        thickness : float or ndarray
            Thickness of waveguide in nm (Valid for 180nm-240nm)
        radius : float or ndarray
            Distance from center of ring to middle of waveguide in nm.
        gap : float or ndarray
            Minimum distance from ring waveguide edge to other ring waveguide edge in nm. (Must be > 100nm)
        sw_angle : float or ndarray, optional
            Sidewall angle of waveguide from horizontal in degrees (Valid for 80-90 degrees). Defaults to 90.
    """

    def __init__(self, width, thickness, radius, gap, sw_angle=90):
        super().__init__(width, thickness, sw_angle)
        self.radius = radius
        self.gap = gap
        if np.any(self.gap < 100):
            warnings.warn(
                "Gap is less than 100nm, may produce invalid results", UserWarning
            )

    def update(self, **kwargs):
        super().update(**kwargs)
        self.radius = kwargs.get("radius", self.radius)
        self.gap = kwargs.get("gap", self.gap)
        if np.any(self.gap < 100):
            warnings.warn(
                "Gap is less than 100nm, may produce invalid results", UserWarning
            )

    def _clean_args(self, wavelength):
        if wavelength is None:
            return clean_inputs(
                (self.width, self.thickness, self.sw_angle, self.radius, self.gap)
            )
        else:
            return clean_inputs(
                (
                    wavelength,
                    self.width,
                    self.thickness,
                    self.sw_angle,
                    self.radius,
                    self.gap,
                )
            )

    def predict(self, ports, wavelength):
        # check to make sure wavelength is in valid range
        if np.any(wavelength < 1450):
            warnings.warn(
                "Wavelength is less than 1450nm, may produce invalid results",
                UserWarning,
            )
        if np.any(wavelength > 1650):
            warnings.warn(
                "Wavelength is greater than 1650nm, may produce invalid results",
                UserWarning,
            )

        # clean data and get coefficients
        wavelength, width, thickness, sw_angle, radius, gap = self._clean_args(
            wavelength
        )
        ae, ao, ge, go, neff = get_coeffs(wavelength, width, thickness, sw_angle)

        # make sure ports are valid
        if not all(1 <= x <= 4 for x in ports):
            raise ValueError("Invalid Ports")

        # determine if cross or through port
        if abs(ports[1] - ports[0]) == 2:
            trig = np.cos
            offset = 0
        else:
            trig = np.sin
            offset = np.pi / 2

        # determine z distance
        if 1 in ports and 3 in ports or 1 in ports and 4 in ports or 2 in ports and 4 in ports or 2 in ports and 3 in ports:
            z_dist = np.pi * radius
        else:
            return np.zeros(len(wavelength))

        # calculate everything
        B = (
            lambda x: 0.5
            * np.pi
            * 2
            * x
            * np.exp(-2 * x)
            * (special.iv(1, 2 * x) + special.modstruve(-1, 2 * x))
        )
        xe = ge * (radius + width / 2)
        xo = go * (radius + width / 2)
        return get_closed_ans(
            ae, ao, ge, go, neff, wavelength, gap, B, xe, xo, offset, trig, z_dist
        )

    def gds(self, filename, extra=0, units="nm", view=False):
        """Writes the geometry to the gds file.

        Parameters
        ----------
        filename : str, optional
            Location to save file to. Defaults to None.
        extra : int, optional
            Extra straight portion to add to ends of waveguides to make room in simulation
            (units same as units parameter). Defaults to 0.
        units : {'microns' or 'nms'}, optional
            Units to save gds file in. Defaults to microns.
        view : bool, optional
            Whether to visually show gds file. Defaults to False.
        """
        raise NotImplementedError("TODO: Write to GDS file")


class AngledHalfRing(DC):
    """This class will create a halfring resonator with a pushed side.

    It takes in a radius and gap along with usual waveguide parameters. Ports are numbered as::

        |      2  \        / 4       |
        |          \      /          |
        |      1--- \    / ---3      |
        |          \ \  / /          |
        |           \ -- /           |
        |            ----            |

    Parameters
    ----------
        width : float or ndarray
            Width of the waveguide in nm (Valid for 400nm-600nm)
        thickness : float or ndarray
            Thickness of waveguide in nm (Valid for 180nm-240nm)
        radius : float or ndarray
            Distance from center of ring to middle of waveguide in nm.
        gap : float or ndarray
            Minimum distance from ring waveguide edge to straight waveguide edge in nm.  (Must be > 100nm)
        theta : float or ndarray
            Angle that the straight waveguide is curved in radians (???).
        sw_angle : float or ndarray, optional (Valid for 80-90 degrees)
    """

    def __init__(self, width, thickness, radius, gap, theta, sw_angle=90):
        super().__init__(width, thickness, sw_angle)
        self.radius = radius
        self.gap = gap
        self.theta = theta
        if np.any(self.gap < 100):
            warnings.warn(
                "Gap is less than 100nm, may produce invalid results", UserWarning
            )

    def update(self, **kwargs):
        super().update(**kwargs)
        self.radius = kwargs.get("radius", self.radius)
        self.gap = kwargs.get("gap", self.gap)
        self.theta = kwargs.get("theta", self.theta)
        if np.any(self.gap < 100):
            warnings.warn(
                "Gap is less than 100nm, may produce invalid results", UserWarning
            )

    def _clean_args(self, wavelength):
        if wavelength is None:
            return clean_inputs(
                (
                    self.width,
                    self.thickness,
                    self.sw_angle,
                    self.radius,
                    self.gap,
                    self.theta,
                )
            )
        else:
            return clean_inputs(
                (
                    wavelength,
                    self.width,
                    self.thickness,
                    self.sw_angle,
                    self.radius,
                    self.gap,
                    self.theta,
                )
            )

    def predict(self, ports, wavelength):
        # check to make sure wavelength is in valid range
        if np.any(wavelength < 1450):
            warnings.warn(
                "Wavelength is less than 1450nm, may produce invalid results",
                UserWarning,
            )
        if np.any(wavelength > 1650):
            warnings.warn(
                "Wavelength is greater than 1650nm, may produce invalid results",
                UserWarning,
            )

        # clean data and get coefficients
        wavelength, width, thickness, sw_angle, radius, gap, theta = self._clean_args(
            wavelength
        )
        ae, ao, ge, go, neff = get_coeffs(wavelength, width, thickness, sw_angle)

        # make sure ports are valid
        if not all(1 <= x <= 4 for x in ports):
            raise ValueError("Invalid Ports")

        # determine if cross or through port
        if abs(ports[1] - ports[0]) == 2:
            trig = np.cos
            offset = 0
        else:
            trig = np.sin
            offset = np.pi / 2

        # determine z distance
        if 1 in ports and 3 in ports:
            z_dist = np.pi * (radius + width + gap)
        elif 1 in ports and 4 in ports or (2 not in ports or 4 not in ports) and 2 in ports and 3 in ports:
            z_dist = np.pi * (radius + width + gap) / 2 + np.pi * radius / 2
        elif 2 in ports and 4 in ports:
            z_dist = np.pi * radius
        else:
            return np.zeros(len(wavelength))

        # calculate everything
        B = lambda x: x
        xe = ge * theta * (radius + width / 2 + gap / 2)
        xo = go * theta * (radius + width / 2 + gap / 2)
        return get_closed_ans(
            ae, ao, ge, go, neff, wavelength, gap, B, xe, xo, offset, trig, z_dist
        )

    def gds(self, filename, extra=0, units="nm", view=False):
        """Writes the geometry to the gds file.

        Parameters
        ----------
        filename : str, optional
            Location to save file to. Defaults to None.
        extra : int, optional
            Extra straight portion to add to ends of waveguides to make room in simulation
            (units same as units parameter). Defaults to 0.
        units : {'microns' or 'nms'}, optional
            Units to save gds file in. Defaults to microns.
        view : bool, optional
            Whether to visually show gds file. Defaults to False.
        """
        raise NotImplementedError("TODO: Write to GDS file")


class Waveguide(ABC):
    """Lossless model for a straight waveguide.

    Simple model that makes sparameters for a straight waveguide. May not be
    the best option, but plays nice with other models in SCEE. Ports are numbered as::

        |  1 ----------- 2   |

    Parameters
    ----------
        width : float or ndarray
            Width of the waveguide in nm (Valid for 400nm-600nm)
        thickness : float or ndarray
            Thickness of waveguide in nm (Valid for 180nm-240nm)
        length : float or ndarray
            Length of waveguide in nm.
        sw_angle : float or ndarray, optional
            Sidewall angle of waveguide from horizontal in degrees (Valid for 80-90 degrees). Defaults to 90.
    """

    def __init__(self, width, thickness, length, sw_angle=90):
        self.width = width
        self.thickness = thickness
        self.length = length
        self.sw_angle = sw_angle
        if np.any(self.width < 400):
            warnings.warn(
                "Width is less than 400nm, may produce invalid results", UserWarning
            )
        if np.any(self.width > 600):
            warnings.warn(
                "Width is greater than 600nm, may produce invalid results", UserWarning
            )
        if np.any(self.thickness < 180):
            warnings.warn(
                "Thickness is less than 180nm, may produce invalid results", UserWarning
            )
        if np.any(self.thickness > 240):
            warnings.warn(
                "Thickness is greater than 240nm, may produce invalid results",
                UserWarning,
            )
        if np.any(self.sw_angle < 80):
            warnings.warn(
                "Sidewall Angle is less than 80 degrees, may produce invalid results",
                UserWarning,
            )
        if np.any(self.sw_angle > 90):
            warnings.warn(
                "Sidewall Angle is greater than 90 degrees, may produce invalid results",
                UserWarning,
            )

    def _clean_args(self, wavelength):
        """Makes sure all attributes are the same size.

        Parses through all self attributes to make sure they're all the same size for
        simulations. Must be reimplemented for all child classes if they have unique attributes.
        Also takes in wavelength parameter to clean as is needed occasionally.

        Parameters
        ----------
        wavelength : float or ndarray
            Wavelength (Valid for 1450nm-1650nm)

        Returns
        ----------
        inputs : (tuple)
            Cleaned array of all devices attributes (and wavelength if included.)
        """
        if wavelength is None:
            return clean_inputs(
                (self.width, self.thickness, self.sw_angle, self.length)
            )
        else:
            return clean_inputs(
                (wavelength, self.width, self.thickness, self.sw_angle, self.length)
            )

    def update(self, **kwargs):
        """Takes in any parameter defined by __init__ and changes it.

        Parameters
        ----------
        attribute : float or ndarray
            Included if any device needs to have an attribute changed.
        """
        self.width = kwargs.get("width", self.width)
        self.thickness = kwargs.get("thickness", self.thickness)
        self.length = kwargs.get("thickness", self.length)
        self.sw_angle = kwargs.get("sw_angle", self.sw_angle)
        if np.any(self.width < 400):
            warnings.warn(
                "Width is less than 400nm, may produce invalid results", UserWarning
            )
        if np.any(self.width > 600):
            warnings.warn(
                "Width is greater than 600nm, may produce invalid results", UserWarning
            )
        if np.any(self.thickness < 180):
            warnings.warn(
                "Thickness is less than 180nm, may produce invalid results", UserWarning
            )
        if np.any(self.thickness > 240):
            warnings.warn(
                "Thickness is greater than 240nm, may produce invalid results",
                UserWarning,
            )
        if np.any(self.sw_angle < 80):
            warnings.warn(
                "Sidewall Angle is less than 80 degrees, may produce invalid results",
                UserWarning,
            )
        if np.any(self.sw_angle > 90):
            warnings.warn(
                "Sidewall Angle is greater than 90 degrees, may produce invalid results",
                UserWarning,
            )

    def sparams(self, wavelength):
        """Returns scattering parameters.

        Runs SCEE to get scattering parameters at wavelength input.

        Parameters
        ----------
        wavelength:    float or ndarray
            wavelengths to get sparams at (Valid for 1450nm-1650nm)

        Returns
        ----------
        s_matrix : ndarray
            size (2,2,n) complex matrix of scattering parameters
        """
        # get number of points to evaluate at
        n = 1 if np.isscalar(wavelength) else len(wavelength)
        # check to make sure the geometry isn't an array
        if len(self._clean_args(None)[0]) != 1:
            raise ValueError(
                "You have changing geometries, getting sparams doesn't make sense"
            )
        s_matrix = np.zeros((2, 2, n), dtype="complex")

        # calculate upper half of matrix (diagonal is 0)
        s_matrix[0, 1] = self.predict(wavelength)

        # apply symmetry (note diagonal is 0, no need to subtract it)
        s_matrix += np.transpose(s_matrix, (1, 0, 2))

        # transpose so depth comes first
        s_matrix = np.transpose(s_matrix, (2, 0, 1))
        return s_matrix

    def predict(self, wavelength):
        """Predicts the output when light is put in port 1 and out port 2.

        Parameters
        ----------
        wavelength : float or ndarray
            Wavelength(s) to predict at (Valid for 1450nm-1650nm)

        Returns
        ----------
        k/t : complex ndarray
            The value of the light coming through
        """
        # check to make sure wavelength is in valid range
        if np.any(wavelength < 1450):
            warnings.warn(
                "Wavelength is less than 1450nm, may produce invalid results",
                UserWarning,
            )
        if np.any(wavelength > 1650):
            warnings.warn(
                "Wavelength is greater than 1650nm, may produce invalid results",
                UserWarning,
            )

        # clean data and get coefficients
        wavelength, width, thickness, sw_angle, length = self._clean_args(wavelength)
        ae, ao, ge, go, neff = get_coeffs(wavelength, width, thickness, sw_angle)

        # calculate everything
        z_dist = length
        phase = 2 * z_dist * neff * np.pi / wavelength

        return np.exp(-1j * phase)

    def gds(self, filename=None, extra=0, units="microns", view=False):
        """Writes the geometry to the gds file.

        Parameters
        ----------
            filename : str
                location to save file to, or if you don't want to defaults to None
            extra : int
                extra straight portion to add to ends of waveguides to make room in simulation
                                (input with units same as units input)
            units : str
                either 'microns' or 'nms'. Units to save gds file in
        """
        # check to make sure the geometry isn't an array
        if len(self._clean_args(None)[0]) != 1:
            raise ValueError(
                "You have changing geometries, making gds doesn't make sense"
            )

        if units == "nms":
            scale = 1
        elif units == "microns":
            scale = 10 ** -3
        else:
            raise ValueError("Invalid units")

        # scale to proper units
        sc_width = self.width * scale
        sc_length = self.length * scale

        # write to GDS
        path = gdspy.Path(sc_width, (-sc_length / 2 - extra, 0))
        path.segment(2 * extra + sc_length, "+x")

        gdspy.current_library = gdspy.GdsLibrary()
        path_cell = gdspy.Cell("C0")
        path_cell.add(path)

        if view:
            gdspy.LayoutViewer(cells=path_cell)

        if filename is not None:
            writer = gdspy.GdsWriter(filename, unit=1.0e-6, precision=1.0e-9)
            writer.write_cell(path_cell)
            writer.close()



import numpy as np
from simphony import Model
from simphony.tools import freq2wl
from simphony.tools import wl2freq

########################################################################
####### This file contains integration of SCEE into various other ######
#######    tools, allowing for cascading of complex structures    ######
########################################################################


def export_interconnect(sparams, wavelength, filename, clear=True):
    """Exports scattering parameters to a file readable by interconnect.

    Parameters
    -----------
    sparams : ndarray
        Numpy array of size (N, d, d) where N is the number of frequency points and d the number of ports
    wavelength : ndarray
        Numpy array of wavelengths (in nm, like the rest of SCEE) of size (N)
    filename : string
        Location to save file
    clear : bool, optional
        If True, empties the file first. Defaults to True.
    """
    # set things up
    _, d, _ = sparams.shape
    if clear:
        open(filename, "w").close()
    with open(filename, "ab") as file:
        # make frequencies
        freq = wl2freq(wavelength * 1e-9)

        # iterate through sparams saving
        for in_ in range(d):
            for out in range(d):
                # put things together
                sp = sparams[:, in_, out]
                temp = np.vstack((freq, np.abs(sp), np.unwrap(np.angle(sp)))).T

                # Save header
                header = f'("port {out+1}", "TE", 1, "port {in_+1}", 1, "transmission")\n'
                header += f"{temp.shape}"

                # save data
                np.savetxt(file, temp, header=header, comments="")


class SimphonyWrapper(Model):
    """Class that wraps SCEE models for use in simphony.

    Model passed into class CANNOT have varying geometries, as a device such as this
    can't be cascaded properly.

    Parameters
    -----------
    model : DC
        Chosen compact model from ``SiPANN.scee`` module. Can be any model that inherits from
        the DC abstract class
    sigmas : dict, optional
        Dictionary mapping parameters to sigma values for use in monte_carlo simulations. Note sigmas should
        be in values of nm. Defaults to an empty dictionary.
    """

    pins = ("n1", "n2", "n3", "n4")  #: The default pin names of the device
    freq_range = (
        182800279268292.0,
        205337300000000.0,
    )  #: The valid frequency range for this model.

    def __init__(self, model, sigmas=dict()):
        self.model = model
        self.sigmas = sigmas

        # save actual parameters for switching back from monte_carlo
        self.og_params = self.model.__dict__.copy()
        self.rand_params = {}

        # make sure there's no varying geometries
        args = self.model._clean_args(None)
        if len(args[0]) != 1:
            raise ValueError(
                "You have changing geometries, use in simphony doesn't make sense!"
            )

        self.regenerate_monte_carlo_parameters()

    def s_parameters(self, freq):
        """Get the s-parameters of SCEE Model.

        Parameters
        ----------
        freq : np.ndarray
            A frequency array to calculate s-parameters over (in Hz).

        Returns
        -------
        s : np.ndarray
            Returns the calculated s-parameter matrix.
        """
        # convert wavelength to frequency
        wl = freq2wl(freq) * 1e9

        return self.model.sparams(wl)

    def monte_carlo_s_parameters(self, freq):
        """Get the s-parameters of SCEE Model with slightly changed parameters.

        Parameters
        ----------
        freq : np.ndarray
            A frequency array to calculate s-parameters over (in Hz).

        Returns
        -------
        s : np.ndarray
            Returns the calculated s-parameter matrix.
        """
        wl = freq2wl(freq) * 1e9

        # perturb params and get sparams
        self.model.update(**self.rand_params)
        sparams = self.model.sparams(wl)

        # restore parameters to originals
        self.model.update(**self.og_params)

        return sparams

    def regenerate_monte_carlo_parameters(self):
        """Varies parameters based on passed in sigma dictionary.

        Iterates through sigma dictionary to change each of those
        parameters, with the mean being the original values found in
        model.
        """
        # iterate through all params that should be tweaked
        for param, sigma in self.sigmas.items():
            self.rand_params[param] = np.random.normal(self.og_params[param], sigma)
